def in_control(data, UCL, CL, LCL, wel=False):
    """Kontrolliert welche Punkte nicht unter kontrolle sind.

    Args:
        data: Stichproben Mittelwerte oder Spannweiten
        UCL: Obere Kontrollgrenze
        CL: Mittellinie
        LCL: Untere Kontrollgrenze
        wel: Flag ob Western Electric Rules mit kontrolliert werden sollen
    """
    # Speichert Punkte welche nicht unter Kontrolle sind
    out_of_control = []

    # Überprüfen, ob jede Stichprobe innerhalb der Grenzen liegt
    for i, point in enumerate(data):
        if point > UCL or point < LCL:
            out_of_control.append(i)  # Speichern der Stichprobe (Zeilennummer)

    # Punkte ausser Kontrolle in Konsole schreiben
    if out_of_control:
        print(f"Die folgenden punkte sind ausser Kontrolle: {out_of_control}")
    else:
        print("Alle Stichproben sind unter Kontrolle.")

    return out_of_control



from lost_spc.calculations import (
    calculate_control_limits,
    calculate_means,
    calculate_ranges,
    calculate_standard_deviations,
)
from lost_spc.utils import get_sample_size

from .plots import shewhart_card


class R:
    def __init__(self, data, z=3, plot_calibration_data=True, restrict_zero=True):
        sample_size = get_sample_size(data)
        self.z = z
        self.m = sample_size.m
        self.plot_calibration_data = plot_calibration_data
        self.restrict_zero = restrict_zero

    def fit(self, X):
        control_limits = calculate_control_limits(X, chart_type="R", z=self.z)
        self.UCL = control_limits["UCL"]
        self.CL = control_limits["CL"]
        self.LCL = control_limits["LCL"]
        if self.LCL < 0 and self.restrict_zero:
            self.LCL = 0
        if self.plot_calibration_data:
            self.R_i_cal = calculate_ranges(X)

    def transform(self, X):
        if self.plot_calibration_data:
            cal_samples = self.R_i_cal
        else:
            cal_samples = None
        R_i = calculate_ranges(X)
        fig = shewhart_card(
            self.UCL,
            self.CL,
            self.LCL,
            R_i,
            calibration_samples=cal_samples,
            title=r"$R$-Karte",
            ylabel=r"$R_i$",
            restrict_zero=self.restrict_zero,
        )
        return fig


class X_R:
    def __init__(self, data, z=3, plot_calibration_data=True, restrict_zero=True):
        sample_size = get_sample_size(data)
        self.z = z
        self.m = sample_size.m
        self.plot_calibration_data = plot_calibration_data
        self.restrict_zero = restrict_zero

    def fit(self, X):
        control_limits = calculate_control_limits(X, chart_type="X_R", z=self.z)
        self.UCL = control_limits["UCL"]
        self.CL = control_limits["CL"]
        self.LCL = control_limits["LCL"]
        if self.LCL < 0 and self.restrict_zero:
            self.LCL = 0
        if self.plot_calibration_data:
            self.X_mean_cal = calculate_means(X)

    def transform(self, X):
        if self.plot_calibration_data:
            cal_samples = self.X_mean_cal
        else:
            cal_samples = None
        X_mean = calculate_means(X)
        fig = shewhart_card(
            self.UCL,
            self.CL,
            self.LCL,
            X_mean,
            calibration_samples=cal_samples,
            title=r"$\bar{X}$-Karte",
            ylabel=r"$\bar{X}_i$",
            restrict_zero=self.restrict_zero,
        )
        return fig


class S:
    def __init__(self, data, z=3, plot_calibration_data=True, restrict_zero=True):
        sample_size = get_sample_size(data)
        self.z = z
        self.m = sample_size.m
        self.plot_calibration_data = plot_calibration_data
        self.restrict_zero = restrict_zero

    def fit(self, X):
        control_limits = calculate_control_limits(X, chart_type="S", z=self.z)
        self.UCL = control_limits["UCL"]
        self.CL = control_limits["CL"]
        self.LCL = control_limits["LCL"]
        if self.LCL < 0 and self.restrict_zero:
            self.LCL = 0
        if self.plot_calibration_data:
            self.S_i_cal = calculate_standard_deviations(X)

    def transform(self, X):
        if self.plot_calibration_data:
            cal_samples = self.S_i_cal
        else:
            cal_samples = None
        S_i = calculate_standard_deviations(X)
        fig = shewhart_card(
            self.UCL,
            self.CL,
            self.LCL,
            S_i,
            calibration_samples=cal_samples,
            title=r"$S$-Karte",
            ylabel=r"$S_i$",
            restrict_zero=self.restrict_zero,
        )
        return fig


class X_S:
    def __init__(self, data, z=3, plot_calibration_data=False, restrict_zero=True):
        sample_size = get_sample_size(data)
        self.z = z
        self.m = sample_size.m
        self.plot_calibration_data = plot_calibration_data
        self.restrict_zero = restrict_zero

    def fit(self, X):
        control_limits = calculate_control_limits(X, chart_type="X_S", z=self.z)
        self.UCL = control_limits["UCL"]
        self.CL = control_limits["CL"]
        self.LCL = control_limits["LCL"]
        if self.LCL < 0 and self.restrict_zero:
            self.LCL = 0
        if self.plot_calibration_data:
            self.X_mean_cal = calculate_means(X)

    def transform(self, X):
        if self.plot_calibration_data:
            cal_samples = self.X_mean_cal
        else:
            cal_samples = None
        fig = X_mean = calculate_means(X)
        shewhart_card(
            self.UCL,
            self.CL,
            self.LCL,
            X_mean,
            calibration_samples=cal_samples,
            title=r"$\bar{X}$-Karte",
            ylabel=r"$\bar{X}_i$",
            restrict_zero=self.restrict_zero,
        )
        return fig



import numpy as np
import scipy.stats as st
from scipy.special import gamma

from ._d_table import _D_TABLE, D


def _check_sample_size(m: int):
    if m < 2:
        raise ValueError("The sample size m has to be >= 2.")


def get_d(m: int, sim_size: int = 10_000) -> D:
    """Correction factors for the R and X control charts.

    Refer to
    [Shewhart charts](http://spc.pages.gitlab.ost.ch/statistische-\
    qualitaetskontrolle/cc_variable_data.html#die-shewhart-kontrollkarte) for
    further reference.

    Args:
        m: Size of each sample.
        sim_size: Number of simulations performed to estimate d2 and d3 if no
                  predifined value is available.

    Returns:
        Dataclass with the corrisponding d2 and d3 value.

    Examples:
        >>> d = get_d(5)
        >>> d.d2
        2.326
        >>> d.d3
        0.864
    """
    _check_sample_size(m)

    if m in _D_TABLE:
        return _D_TABLE[m]

    x = np.array(st.norm.rvs(size=(sim_size, m)))
    R_i = x.max(axis=1) - x.min(axis=1)
    d2 = np.mean(R_i)
    d3 = np.std(R_i, ddof=1)
    return D(d2=d2, d3=d3)


def get_c4(m: int) -> float:
    """Correction factor used for the s and corrisponding X control charts.

    Refer to
    http://spc.pages.gitlab.ost.ch/statistische-qualitaetskontrolle/cc_variable_data.html#s-karte
    for further reference.

    Args:
        m: Size of each sample.

    Returns:
        Calculated c4 value.

    Examples:
        >>> get_c4(5)
        np.float64(0.9399856029866253)
    """
    _check_sample_size(m)

    return gamma(m / 2) / gamma((m - 1) / 2) * np.sqrt(2 / (m - 1))



from .control_limits import calculate_control_limits
from .spc_values import calculate_means, calculate_ranges, calculate_standard_deviations

__all__ = [
    "calculate_control_limits",
    "calculate_means",
    "calculate_ranges",
    "calculate_standard_deviations",
]



import matplotlib.pyplot as plt


def shewhart_card(
    UCL,
    CL,
    LCL,
    samples,
    calibration_samples=None,
    title="",
    ylabel="",
    fill_alpha=0.07,
    restrict_zero=True,
):
    """
    Plots a shewhart_card.

    restict_zero: Switch to restrict the lower control limit to 0 if it's below.
    """
    if restrict_zero:
        if LCL < 0:
            LCL = 0

    if calibration_samples is not None:
        x_min = -len(calibration_samples)
    else:
        x_min = 0

    # Draw control limits
    plt.hlines([UCL, CL, LCL], x_min, len(samples) - 1, colors=["black"], alpha=0.8)
    area_height = (UCL - CL) / 3
    plt.hlines(
        [CL + area_height, CL + 2 * area_height, CL - area_height, CL - 2 * area_height],
        x_min,
        len(samples) - 1,
        colors=["black"],
        alpha=0.3,
        linestyles="dashed",
    )

    # Add some coloring for the areas
    width = (x_min, len(samples) - 1)
    plt.fill_between(width, CL - area_height, CL + area_height, alpha=fill_alpha, color="green")
    plt.fill_between(
        width, CL + area_height, CL + 2 * area_height, alpha=fill_alpha, color="yellow"
    )
    plt.fill_between(
        width, CL - area_height, CL - 2 * area_height, alpha=fill_alpha, color="yellow"
    )
    plt.fill_between(width, CL + 2 * area_height, UCL, alpha=fill_alpha, color="red")
    plt.fill_between(width, CL - 2 * area_height, LCL, alpha=fill_alpha, color="red")

    # Plot points
    if calibration_samples is not None:
        plt.vlines(0, ymin=LCL, ymax=UCL, colors=["red"], linestyles="dotted", alpha=0.6)
        plt.plot(range(-len(calibration_samples), 0, 1), calibration_samples, "o-")
    plt.plot(range(len(samples)), samples, "o-")

    # Plot setup
    plt.title(title)
    plt.xlabel("Sample")
    plt.ylabel(ylabel)
    plt.grid()
    return plt.gcf()



import numpy as np

from lost_spc.constants import get_c4, get_d
from lost_spc.utils import get_sample_size

from .spc_values import calculate_means, calculate_ranges, calculate_standard_deviations


def calculate_control_limits(data: np.ndarray, chart_type: str = "X_R", z: int = 3) -> dict:
    """Calculates the control limits for different SPC charts (X̄, R, S, etc.).

    Args:
        data (np.ndarray): The data for which the control limits are to be calculated.
        chart_type (str): Type of chart ('X_R', 'R', 'S' or 'X_S'). Determines
                          the control limits' calculations.
        z (int): The number of standard deviations for control limits. Default is 3.

    Returns:
        dict: A dictionary containing the center line (CL), upper control limit (UCL),
              and lower control limit (LCL).

    Examples:
        >>> data = np.array([[10, 12, 14], [15, 16, 17]])
        >>> calculate_control_limits(data, chart_type='X_R')
        {'CL': 14.0, 'UCL': 17.069198123276216, 'LCL': 10.930801876723784}
    """
    # Bestimme m und n automatisch für NumPy-Arrays
    sample_size = get_sample_size(data)
    m = sample_size.m  # Stichprobengröße

    # Berechne die Kontrollgrenzen basierend auf dem Kartentyp
    if chart_type == "X_R":
        # X̄-R Karte: Mittelwert und Spannweite
        means = calculate_means(data)
        ranges = calculate_ranges(data)
        d = get_d(m)
        d2 = d.d2
        cl = np.mean(means)
        R_mean = np.mean(ranges)
        factor = z * (R_mean / d2) / np.sqrt(m)
        ucl = cl + factor
        lcl = cl - factor
    elif chart_type == "R":
        # R-Karte: Spannweite
        ranges = calculate_ranges(data)
        d = get_d(m)
        d2 = d.d2
        d3 = d.d3
        range_mean = np.mean(ranges)
        cl = range_mean
        factor = cl * z * d3 / d2
        ucl = cl + factor
        lcl = cl - factor
    elif chart_type == "S":
        # S-Karte: Standardabweichung
        std_devs = calculate_standard_deviations(data)
        c4 = get_c4(m)
        cl = np.mean(std_devs)
        factor = cl * z * np.sqrt(1 - c4**2) / c4
        ucl = cl + factor
        lcl = cl - factor
    elif chart_type == "X_S":
        # X̄-S Karte: Mittelwert und Standardabweichung
        means = calculate_means(data)
        std_devs = calculate_standard_deviations(data)
        X_mean = np.mean(means)
        s_i = std_devs
        s_mean = np.mean(s_i)
        c4 = get_c4(m)
        cl = X_mean
        factor = z * (s_mean / c4) / np.sqrt(m)
        ucl = cl + factor
        lcl = cl - factor
    else:
        raise ValueError(f"Unsupported chart type: {chart_type}")

    return {"CL": float(cl), "UCL": float(ucl), "LCL": float(lcl)}



from dataclasses import dataclass

import numpy as np


@dataclass
class SampleSize:
    m: int  # Stichprobengrösse
    n: int  # Anzahl Stichproben


def get_sample_size(data: np.ndarray) -> SampleSize:
    """Returns the sample size (number of columns) and the number of rows (probes)
    for a NumPy array dataset.

    Args:
        data (np.ndarray): 2D NumPy array of numerical values.

    Returns:
        SampleSize: A dataclass containing the sample size (m) and the number of probes (n).

    Examples:
        >>> data = np.array([[10, 12, 14], [15, 16, 17]])
        >>> get_sample_size(data)
        SampleSize(m=3, n=2)
    """
    if isinstance(data, np.ndarray):
        m = data.shape[1]  # Stichprobengrösse
        n = data.shape[0]  # Anzahl Stichproben
        return SampleSize(m=m, n=n)
    else:
        raise TypeError("Input data must be a NumPy array.")



import numpy as np


def calculate_means(data: np.ndarray) -> np.ndarray:
    """Calculates the mean (average) of each row (sample) in a dataset.

    Args:
        data (np.ndarray): Array of numerical values.

    Returns:
        np.ndarray: Array of means for each row.

    Examples:
        >>> calculate_means(np.array([[10, 12, 14], [15, 16, 17]]))
        array([12., 16.])
    """
    return np.mean(data, axis=1)


def calculate_ranges(data: np.ndarray) -> np.ndarray:
    """Calculates the range (max - min) of each row (sample) in a dataset.

    Args:
        data (np.ndarray): Array of numerical values.

    Returns:
        np.ndarray: Array of ranges for each row.

    Examples:
        >>> calculate_ranges(np.array([[10, 12, 14], [15, 16, 17]]))
        array([4., 2.])
    """
    return np.ptp(data, axis=1).astype(float)


def calculate_standard_deviations(data: np.ndarray) -> np.ndarray:
    """Calculates the sample standard deviation for each row (sample) in the dataset.

    Args:
        data (np.ndarray): Array of numerical values, where each row is a sample.

    Returns:
        np.ndarray: Array of sample standard deviations for each row.

    Examples:
        >>> calculate_standard_deviations(np.array([[10, 12, 14], [15, 16, 17]]))
        array([2., 1.])
    """
    return np.std(data, axis=1, ddof=1)



from dataclasses import dataclass
from typing import Union

from numpy import floating


@dataclass
class D:
    d2: Union[float, floating]
    d3: Union[float, floating]


# Source: https://www.qimacros.com/control-chart/control-chart-constants/
# Checked with: Introduction to Statistical Quality Control, Douglas C. Montgomery
#               ISBN: 978-1-118-14681-1
_D_TABLE = {
    2: D(d2=1.128, d3=0.853),
    3: D(d2=1.693, d3=0.888),
    4: D(d2=2.059, d3=0.880),
    5: D(d2=2.326, d3=0.864),
    6: D(d2=2.534, d3=0.848),
    7: D(d2=2.704, d3=0.833),
    8: D(d2=2.847, d3=0.820),
    9: D(d2=2.970, d3=0.808),
    10: D(d2=3.078, d3=0.797),
    15: D(d2=3.472, d3=0.756),
    25: D(d2=3.931, d3=0.708),
}
